==> illumination_0.1.20090329-1.tar.gz: illumination-0.1.20090329/src/Lum/OS/Hildon/KeyboardManager.cpp <==
/*
  This source is part of the Illumination library
  Copyright (C) 2008  Tim Teulings

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
*/

#include <Lum/OS/Hildon/KeyboardManager.h>

#include <X11/Xatom.h>

#include <hildon-input-method/hildon-im-protocol.h>
#include <gtk/gtkenums.h>

#include <Lum/Base/String.h>

#include <Lum/OS/X11/Display.h>
#include <Lum/OS/X11/Event.h>
#include <Lum/OS/X11/Window.h>

#include <Lum/Window.h>

#include <iostream>

namespace Lum {
  namespace OS {
    namespace Hildon {

      KeyboardManager::KeyboardManager()
      : window(NULL),
        object(NULL)
      {
        // no code
      }

      KeyboardManager::~KeyboardManager()
      {
        // no code
      }

      void KeyboardManager::Shutdown()
      {
        HideKeyboard();
      }

      void KeyboardManager::OnWindowFocusIn(OS::Window* window)
      {
        this->window=window;
        this->object=NULL;
      }

      void KeyboardManager::OnWindowFocusOut(OS::Window* window)
      {
        // no code
      }

      void KeyboardManager::OnFocusChange(OS::Window* window,
                                          Lum::Object* oldObject,
                                          Lum::Object* newObject)
      {
        if (this->window!=window) {
          return;
        }

        if (object==newObject) {
          return;
        }

        if (object!=NULL && object->RequestsKeyboard() &&
            newObject!=NULL && newObject->RequestsKeyboard()) { // "do nothing" case
          object=newObject;
        }
        else if (newObject!=NULL && newObject->RequestsKeyboard()) { // "show" case
          object=newObject;
          ShowKeyboard();
        }
        else { // "hide" case
          object=NULL;
          HideKeyboard();
        }
      }

      bool KeyboardManager::HandleEvent(OS::Window* window, OS::Event* event)
      {
        XEvent x11Event;

        X11::GetX11Event(event,x11Event);

        if (x11Event.type!=ClientMessage) {
          return false;
        }

        if (this->window!=window) {
          return false;
        }

        X11::Display *display=dynamic_cast<X11::Display*>(Lum::OS::display);

        if (display==NULL) {
          return false;
        }

        ::Window imWindow=GetKeyboardWindow();

        if (imWindow==None) {
          return false;
        }

        if (x11Event.xclient.message_type==XInternAtom(display->display,HILDON_IM_COM_NAME,false)) {
          HildonIMComMessage *msg = (HildonIMComMessage *)&x11Event.xclient.data;

          if (msg->type==HILDON_IM_CONTEXT_HANDLE_ENTER) {
            X11::KeyEvent keyEvent;

            keyEvent.key=keyReturn;
            keyEvent.qualifier=0;
            keyEvent.type=KeyEvent::down;

            window->HandleEvent(&keyEvent);

            keyEvent.type=KeyEvent::up;

            window->HandleEvent(&keyEvent);
            return true;
          }
          else if (msg->type==HILDON_IM_CONTEXT_HANDLE_TAB) {
            X11::KeyEvent keyEvent;

            keyEvent.key=keyTab;
            keyEvent.qualifier=0;
            keyEvent.type=KeyEvent::down;

            window->HandleEvent(&keyEvent);

            keyEvent.type=KeyEvent::up;

            window->HandleEvent(&keyEvent);
            return true;
          }
          else if (msg->type==HILDON_IM_CONTEXT_HANDLE_BACKSPACE) {
            X11::KeyEvent keyEvent;

            keyEvent.key=keyBackspace;
            keyEvent.qualifier=0;
            keyEvent.type=KeyEvent::down;

            window->HandleEvent(&keyEvent);

            keyEvent.type=KeyEvent::up;

            window->HandleEvent(&keyEvent);
            return true;
          }
          else if (msg->type==HILDON_IM_CONTEXT_HANDLE_SPACE) {
            X11::KeyEvent keyEvent;

            keyEvent.key=keySpace;
            keyEvent.qualifier=0;
            keyEvent.type=KeyEvent::down;

            window->HandleEvent(&keyEvent);

            keyEvent.type=KeyEvent::up;

            window->HandleEvent(&keyEvent);
            return true;
          }
          else if (msg->type==HILDON_IM_CONTEXT_CONFIRM_SENTENCE_START) {
            //std::cout << "HILDON_IM_CONTEXT_CONFIRM_SENTENCE_START" << std::endl;

            ::XEvent event;

            memset(&event,0,sizeof(event));
            event.xclient.type=ClientMessage;
            event.xclient.window=imWindow;
            event.xclient.message_type=XInternAtom(display->display,HILDON_IM_ACTIVATE_NAME,False);
            event.xclient.format=8;
            event.xclient.data.l[0]=dynamic_cast<X11::Window*>(window)->GetDrawable();
            event.xclient.data.l[1]=dynamic_cast<X11::Window*>(window)->GetDrawable();
            event.xclient.data.l[2]=0x03;
            event.xclient.data.l[3]=0x00000007;
            event.xclient.data.l[4]=0;
            if (XSendEvent(display->display,imWindow,False,NoEventMask,&event)==0) {
              std::cerr << "Error while sending _HILDON_IM_ACTIVATE" << std::endl;
            }
          }
          else if (msg->type==HILDON_IM_CONTEXT_FLUSH_PREEDIT) {
            //std::cout << "HILDON_IM_CONTEXT_FLUSH_PREEDIT" << std::endl;
          }
          else if (msg->type==HILDON_IM_CONTEXT_BUFFERED_MODE) {
            //std::cout << "HILDON_IM_CONTEXT_BUFFERED_MODE" << std::endl;
          }
          else if (msg->type==HILDON_IM_CONTEXT_DIRECT_MODE) {
            //std::cout << "HILDON_IM_CONTEXT_DIRECT_MODE" << std::endl;
          }
          else if (msg->type==HILDON_IM_CONTEXT_REDIRECT_MODE) {
            //std::cout << "HILDON_IM_CONTEXT_REDIRECT_MODE" << std::endl;
          }
          else if (msg->type==HILDON_IM_CONTEXT_SURROUNDING_MODE) {
            //std::cout << "HILDON_IM_CONTEXT_SURROUNDING_MODE" << std::endl;
          }
          else if (msg->type==HILDON_IM_CONTEXT_CLIPBOARD_COPY) {
            //std::cout << "HILDON_IM_CONTEXT_CLIPBOARD_COPY" << std::endl;
            Copy();
          }
          else if (msg->type==HILDON_IM_CONTEXT_CLIPBOARD_CUT) {
            //std::cout << "HILDON_IM_CONTEXT_CLIPBOARD_CUT" << std::endl;
            Cut();
          }
          else if (msg->type==HILDON_IM_CONTEXT_CLIPBOARD_PASTE) {
            //std::cout << "HILDON_IM_CONTEXT_CLIPBOARD_PASTE" << std::endl;
            Paste();
          }
          else if (msg->type==HILDON_IM_CONTEXT_CLIPBOARD_SELECTION_QUERY) {
            //std::cout << "HILDON_IM_CONTEXT_CLIPBOARD_SELECTION_QUERY" << std::endl;

            ::XEvent                event;
            bool                    hasSelection;

            hasSelection=object!=NULL && object->CanExecuteAction(Lum::Base::Datatype(Lum::Base::Datatype::text,
                                                                                      Lum::Base::Datatype::plain),
                                                                  Lum::Base::DataExchangeObject::actionCopy);

            std::cout << "HasSelection: " << hasSelection << std::endl;

            memset(&event,0,sizeof(event));
            event.xclient.type=ClientMessage;
            event.xclient.window=imWindow;
            event.xclient.message_type=XInternAtom(display->display,HILDON_IM_CLIPBOARD_SELECTION_REPLY_NAME,False);
            event.xclient.format=HILDON_IM_CLIPBOARD_SELECTION_REPLY_FORMAT;
            event.xclient.data.l[0]=hasSelection;

            if (XSendEvent(display->display,imWindow,False,NoEventMask,&event)==0) {
              std::cerr << "Error while sending HILDON_IM_CLIPBOARD_SELECTION_REPLY" << std::endl;
            }
          }
          else if (msg->type==HILDON_IM_CONTEXT_REQUEST_SURROUNDING) {
            //std::cout << "HILDON_IM_CONTEXT_REQUEST_SURROUNDING" << std::endl;
          }
          else if (msg->type==HILDON_IM_CONTEXT_WIDGET_CHANGED) {
            //std::cout << "HILDON_IM_CONTEXT_WIDGET_CHANGED" << std::endl;
          }
          else if (msg->type==HILDON_IM_CONTEXT_OPTION_CHANGED) {
            //std::cout << "HILDON_IM_CONTEXT_OPTION_CHANGED" << std::endl;
          }
          else if (msg->type==HILDON_IM_CONTEXT_CLEAR_STICKY) {
            //std::cout << "HILDON_IM_CONTEXT_CLEAR_STICKY" << std::endl;
          }
          //else if (msg->type==HILDON_IM_CONTEXT_ENTER_ON_FOCUS) {
            //std::cout << "HILDON_IM_CONTEXT_ENTER_ON_FOCUS" << std::endl;
          //}
          else {
            std::cout << "got HILDON-COM " << std::hex << msg->type << std::endl;
          }
        }
        else if (x11Event.xclient.message_type==XInternAtom(display->display,HILDON_IM_INSERT_UTF8_NAME,false)) {
          X11::KeyEvent keyEvent;
          std::string   inputText;
          size_t        i;

          // Decode text
          i=4;
          while (i<20 && x11Event.xclient.data.b[i]!='\0') {
            inputText.append(1,(char)x11Event.xclient.data.b[i]);
            i++;
          }

          // Fake key event
          keyEvent.key=keyUnknown;
          keyEvent.qualifier=0;
          keyEvent.text=Lum::Base::UTF8ToWString(inputText);
          keyEvent.type=KeyEvent::down;

          window->HandleEvent(&keyEvent);

          keyEvent.type=KeyEvent::up;

          window->HandleEvent(&keyEvent);
        }
        else if (x11Event.xclient.message_type==XInternAtom(display->display,HILDON_IM_SURROUNDING_CONTENT_NAME,false)) {
          //std::cout << HILDON_IM_SURROUNDING_CONTENT_NAME << std::endl;
        }
        else if (x11Event.xclient.message_type==XInternAtom(display->display,HILDON_IM_SURROUNDING_NAME,false)) {
          //std::cout << HILDON_IM_SURROUNDING_NAME << std::endl;
        }

        return false;
      }

      void KeyboardManager::ShowKeyboard()
      {
        assert(window!=0);

        X11::Display *display=dynamic_cast<X11::Display*>(Lum::OS::display);

        if (display==NULL) {
          return;
        }

        ::Window imWindow=GetKeyboardWindow();

        if (imWindow==None) {
          return;
        }

        ::XEvent event;

        memset(&event,0,sizeof(event));
        event.xclient.type=ClientMessage;
        event.xclient.window=imWindow;
        event.xclient.message_type=XInternAtom(display->display,HILDON_IM_ACTIVATE_NAME,False);
        event.xclient.format=HILDON_IM_ACTIVATE_FORMAT;

        HildonIMActivateMessage *msg = reinterpret_cast<HildonIMActivateMessage *>(&event.xclient.data);

        msg->input_window=dynamic_cast<X11::Window*>(window)->GetDrawable();
        msg->app_window=dynamic_cast<X11::Window*>(window)->GetDrawable();
        msg->cmd=HILDON_IM_SETNSHOW;
        msg->input_mode=HILDON_GTK_INPUT_MODE_FULL;
        msg->trigger=HILDON_IM_TRIGGER_STYLUS;

        if (XSendEvent(display->display,imWindow,False,NoEventMask,&event)==0) {
          std::cerr << "Error while sending OpenKeyboard (_HILDON_IM_ACTIVATE)" << std::endl;
        }

        XSync(display->display,false);
      }

      void KeyboardManager::HideKeyboard()
      {
        X11::Display *display=dynamic_cast<X11::Display*>(Lum::OS::display);

        if (display==NULL) {
          return;
        }

        ::Window imWindow=GetKeyboardWindow();

        if (imWindow==None) {
          return;
        }

        ::XEvent event;

        memset(&event,0,sizeof(event));
        event.xclient.type=ClientMessage;
        event.xclient.window=imWindow;
        event.xclient.message_type=XInternAtom(display->display,HILDON_IM_ACTIVATE_NAME,False);
        event.xclient.format=HILDON_IM_ACTIVATE_FORMAT;

        HildonIMActivateMessage *msg = reinterpret_cast<HildonIMActivateMessage *>(&event.xclient.data);

        msg->input_window=0;
        msg->app_window=0;
        msg->cmd=HILDON_IM_HIDE;
        msg->input_mode=HILDON_GTK_INPUT_MODE_FULL;
        msg->trigger=HILDON_IM_TRIGGER_STYLUS;

        if (XSendEvent(display->display,imWindow,False,NoEventMask,&event)==0) {
          std::cerr << "Error while sending 'CloseKeyboard' (_HILDON_IM_ACTIVATE)" << std::endl;
        }

        XSync(display->display,false);
      }

      ::Window KeyboardManager::GetKeyboardWindow() const
      {
        X11::Display *display=dynamic_cast<X11::Display*>(Lum::OS::display);

        if (display==NULL) {
          return None;
        }

        Atom hildonWindow=XInternAtom(display->display,"_HILDON_IM_WINDOW",true);

        if (hildonWindow==None) {
          return None;
        }

        Atom           atomType;
        int            atomFormat;
        int            res;
        unsigned long  atomItems;
        unsigned long  atomBytes;
        unsigned char* atomProperties;
        ::Window       window;

        res=XGetWindowProperty(display->display,RootWindow(display->display,display->scrNum),
                               hildonWindow,0,1,false,display->atoms[display->windowAtom],
                               &atomType,&atomFormat,&atomItems,&atomBytes,&atomProperties);
        if (res!=Success ||
            atomType!=XA_WINDOW ||
            atomFormat!=HILDON_IM_WINDOW_ID_FORMAT ||
            atomItems!=1 ||
            atomProperties==NULL) {
          std::cerr << "Cannot get IM window (_HILDON_IM_WINDOW)" << std::endl;
          return None;
        }

        window=reinterpret_cast< ::Window*>(atomProperties)[0];

        XFree(atomProperties);

        return window;
      }

      void KeyboardManager::Cut()
      {
        Lum::Base::DataExchangeObject *object=OS::display->GetSelectionOwner();

        if (object==NULL) {
          return;
        }

        Lum::Base::Data data;

        if (!object->ExecuteAction(Lum::Base::Datatype(Lum::Base::Datatype::text,
                                                       Lum::Base::Datatype::plain),
                                   data,
                                   Lum::Base::DataExchangeObject::actionMove)) {
          return;
        }

        OS::display->SetClipboard(data.GetText());

        object->PostprocessAction(Lum::Base::DataExchangeObject::actionMove);
      }

      void KeyboardManager::Copy()
      {
        Lum::Base::DataExchangeObject *object=OS::display->GetSelectionOwner();

        if (object==NULL) {
          return;
        }

        Lum::Base::Data data;

        if (!object->ExecuteAction(Lum::Base::Datatype(Lum::Base::Datatype::text,
                                                       Lum::Base::Datatype::plain),
                                   data,
                                   Lum::Base::DataExchangeObject::actionCopy)) {
          return;
        }

        OS::display->SetClipboard(data.GetText());
      }

      void KeyboardManager::Paste()
      {
        if (object==NULL) {
          return;
        }

        std::wstring clipboard;

        clipboard=OS::display->GetClipboard();

        if (clipboard.empty()) {
          return;
        }

        Lum::Base::Data data(clipboard);

        object->ExecuteAction(Lum::Base::Datatype(Lum::Base::Datatype::text,
                                                  Lum::Base::Datatype::plain),
                              data,
                              Lum::Base::DataExchangeObject::actionInsert);
      }
    }
  }
}

==> illumination_0.1.20091009-1.tar.gz: illumination-0.1.20091009/src/Lum/OS/Hildon/KeyboardManager.cpp <==
/*
  This source is part of the Illumination library
  Copyright (C) 2008  Tim Teulings

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
*/

#include <Lum/OS/Hildon/KeyboardManager.h>

#include <X11/Xatom.h>

#include <hildon-input-method/hildon-im-protocol.h>
#include <gtk/gtkenums.h>

#include <Lum/Base/String.h>

#include <Lum/OS/X11/Display.h>
#include <Lum/OS/X11/Event.h>
#include <Lum/OS/X11/Window.h>

#include <Lum/Window.h>

#include <iostream>

namespace Lum {
  namespace OS {
    namespace Hildon {

      KeyboardManager::KeyboardManager()
      : window(NULL),
        object(NULL)
      {
        // no code
      }

      KeyboardManager::~KeyboardManager()
      {
        // no code
      }

      void KeyboardManager::Shutdown()
      {
        HideKeyboard();
      }

      void KeyboardManager::OnWindowFocusIn(OS::Window* window)
      {
        this->window=window;
        this->object=NULL;
      }

      void KeyboardManager::OnWindowFocusOut(OS::Window* window)
      {
        // no code
      }

      void KeyboardManager::OnFocusChange(OS::Window* window,
                                          Lum::Object* oldObject,
                                          Lum::Object* newObject)
      {
        if (this->window!=window) {
          return;
        }

        if (object==newObject) {
          return;
        }

        if (object!=NULL && object->RequestsKeyboard() &&
            newObject!=NULL && newObject->RequestsKeyboard()) { // "do nothing" case
          object=newObject;
        }
        else if (newObject!=NULL && newObject->RequestsKeyboard()) { // "show" case
          object=newObject;
          ShowKeyboard();
        }
        else { // "hide" case
          object=NULL;
          HideKeyboard();
        }
      }

      bool KeyboardManager::HandleEvent(OS::Window* window, OS::Event* event)
      {
        XEvent x11Event;

        X11::GetX11Event(event,x11Event);

        if (x11Event.type!=ClientMessage) {
          return false;
        }

        if (this->window!=window) {
          return false;
        }

        X11::Display *display=dynamic_cast<X11::Display*>(Lum::OS::display);

        if (display==NULL) {
          return false;
        }

        ::Window imWindow=GetKeyboardWindow();

        if (imWindow==None) {
          return false;
        }

        if (x11Event.xclient.message_type==XInternAtom(display->display,HILDON_IM_COM_NAME,false)) {
          HildonIMComMessage *msg = (HildonIMComMessage *)&x11Event.xclient.data;

          if (msg->type==HILDON_IM_CONTEXT_HANDLE_ENTER) {
            X11::KeyEvent keyEvent;

            keyEvent.key=keyReturn;
            keyEvent.qualifier=0;
            keyEvent.type=KeyEvent::down;

            window->HandleEvent(&keyEvent);

            keyEvent.type=KeyEvent::up;

            window->HandleEvent(&keyEvent);
            return true;
          }
          else if (msg->type==HILDON_IM_CONTEXT_HANDLE_TAB) {
            X11::KeyEvent keyEvent;

            keyEvent.key=keyTab;
            keyEvent.qualifier=0;
            keyEvent.type=KeyEvent::down;

            window->HandleEvent(&keyEvent);

            keyEvent.type=KeyEvent::up;

            window->HandleEvent(&keyEvent);
            return true;
          }
          else if (msg->type==HILDON_IM_CONTEXT_HANDLE_BACKSPACE) {
            X11::KeyEvent keyEvent;

            keyEvent.key=keyBackspace;
            keyEvent.qualifier=0;
            keyEvent.type=KeyEvent::down;

            window->HandleEvent(&keyEvent);

            keyEvent.type=KeyEvent::up;

            window->HandleEvent(&keyEvent);
            return true;
          }
          else if (msg->type==HILDON_IM_CONTEXT_HANDLE_SPACE) {
            X11::KeyEvent keyEvent;

            keyEvent.key=keySpace;
            keyEvent.qualifier=0;
            keyEvent.type=KeyEvent::down;

            window->HandleEvent(&keyEvent);

            keyEvent.type=KeyEvent::up;

            window->HandleEvent(&keyEvent);
            return true;
          }
          else if (msg->type==HILDON_IM_CONTEXT_CONFIRM_SENTENCE_START) {
            //std::cout << "HILDON_IM_CONTEXT_CONFIRM_SENTENCE_START" << std::endl;

            ::XEvent event;

            memset(&event,0,sizeof(event));
            event.xclient.type=ClientMessage;
            event.xclient.window=imWindow;
            event.xclient.message_type=XInternAtom(display->display,HILDON_IM_ACTIVATE_NAME,False);
            event.xclient.format=8;
            event.xclient.data.l[0]=dynamic_cast<X11::Window*>(window)->GetDrawable();
            event.xclient.data.l[1]=dynamic_cast<X11::Window*>(window)->GetDrawable();
            event.xclient.data.l[2]=0x03;
            event.xclient.data.l[3]=0x00000007;
            event.xclient.data.l[4]=0;
            if (XSendEvent(display->display,imWindow,False,NoEventMask,&event)==0) {
              std::cerr << "Error while sending _HILDON_IM_ACTIVATE" << std::endl;
            }
          }
          else if (msg->type==HILDON_IM_CONTEXT_FLUSH_PREEDIT) {
            //std::cout << "HILDON_IM_CONTEXT_FLUSH_PREEDIT" << std::endl;
          }
          else if (msg->type==HILDON_IM_CONTEXT_BUFFERED_MODE) {
            //std::cout << "HILDON_IM_CONTEXT_BUFFERED_MODE" << std::endl;
          }
          else if (msg->type==HILDON_IM_CONTEXT_DIRECT_MODE) {
            //std::cout << "HILDON_IM_CONTEXT_DIRECT_MODE" << std::endl;
          }
          else if (msg->type==HILDON_IM_CONTEXT_REDIRECT_MODE) {
            //std::cout << "HILDON_IM_CONTEXT_REDIRECT_MODE" << std::endl;
          }
          else if (msg->type==HILDON_IM_CONTEXT_SURROUNDING_MODE) {
            //std::cout << "HILDON_IM_CONTEXT_SURROUNDING_MODE" << std::endl;
          }
          else if (msg->type==HILDON_IM_CONTEXT_CLIPBOARD_COPY) {
            //std::cout << "HILDON_IM_CONTEXT_CLIPBOARD_COPY" << std::endl;
            Copy();
          }
          else if (msg->type==HILDON_IM_CONTEXT_CLIPBOARD_CUT) {
            //std::cout << "HILDON_IM_CONTEXT_CLIPBOARD_CUT" << std::endl;
            Cut();
          }
          else if (msg->type==HILDON_IM_CONTEXT_CLIPBOARD_PASTE) {
            //std::cout << "HILDON_IM_CONTEXT_CLIPBOARD_PASTE" << std::endl;
            Paste();
          }
          else if (msg->type==HILDON_IM_CONTEXT_CLIPBOARD_SELECTION_QUERY) {
            //std::cout << "HILDON_IM_CONTEXT_CLIPBOARD_SELECTION_QUERY" << std::endl;

            ::XEvent                event;
            bool                    hasSelection;

            hasSelection=object!=NULL && object->CanExecuteAction(Lum::Base::Datatype(Lum::Base::Datatype::text,
                                                                                      Lum::Base::Datatype::plain),
                                                                  Lum::Base::DataExchangeObject::actionCopy);

            std::cout << "HasSelection: " << hasSelection << std::endl;

            memset(&event,0,sizeof(event));
            event.xclient.type=ClientMessage;
            event.xclient.window=imWindow;
            event.xclient.message_type=XInternAtom(display->display,HILDON_IM_CLIPBOARD_SELECTION_REPLY_NAME,False);
            event.xclient.format=HILDON_IM_CLIPBOARD_SELECTION_REPLY_FORMAT;
            event.xclient.data.l[0]=hasSelection;

            if (XSendEvent(display->display,imWindow,False,NoEventMask,&event)==0) {
              std::cerr << "Error while sending HILDON_IM_CLIPBOARD_SELECTION_REPLY" << std::endl;
            }
          }
          else if (msg->type==HILDON_IM_CONTEXT_REQUEST_SURROUNDING) {
            //std::cout << "HILDON_IM_CONTEXT_REQUEST_SURROUNDING" << std::endl;
          }
          else if (msg->type==HILDON_IM_CONTEXT_WIDGET_CHANGED) {
            //std::cout << "HILDON_IM_CONTEXT_WIDGET_CHANGED" << std::endl;
          }
          else if (msg->type==HILDON_IM_CONTEXT_OPTION_CHANGED) {
            //std::cout << "HILDON_IM_CONTEXT_OPTION_CHANGED" << std::endl;
          }
          else if (msg->type==HILDON_IM_CONTEXT_CLEAR_STICKY) {
            //std::cout << "HILDON_IM_CONTEXT_CLEAR_STICKY" << std::endl;
          }
          //else if (msg->type==HILDON_IM_CONTEXT_ENTER_ON_FOCUS) {
            //std::cout << "HILDON_IM_CONTEXT_ENTER_ON_FOCUS" << std::endl;
          //}
          else {
            std::cout << "got HILDON-COM " << std::hex << msg->type << std::endl;
          }
        }
        else if (x11Event.xclient.message_type==XInternAtom(display->display,HILDON_IM_INSERT_UTF8_NAME,false)) {
          X11::KeyEvent keyEvent;
          std::string   inputText;
          size_t        i;

          // Decode text
          i=4;
          while (i<20 && x11Event.xclient.data.b[i]!='\0') {
            inputText.append(1,(char)x11Event.xclient.data.b[i]);
            i++;
          }

          // Fake key event
          keyEvent.key=keyUnknown;
          keyEvent.qualifier=0;
          keyEvent.text=Lum::Base::UTF8ToWString(inputText);
          keyEvent.type=KeyEvent::down;

          window->HandleEvent(&keyEvent);

          keyEvent.type=KeyEvent::up;

          window->HandleEvent(&keyEvent);
        }
        else if (x11Event.xclient.message_type==XInternAtom(display->display,HILDON_IM_SURROUNDING_CONTENT_NAME,false)) {
          //std::cout << HILDON_IM_SURROUNDING_CONTENT_NAME << std::endl;
        }
        else if (x11Event.xclient.message_type==XInternAtom(display->display,HILDON_IM_SURROUNDING_NAME,false)) {
          //std::cout << HILDON_IM_SURROUNDING_NAME << std::endl;
        }

        return false;
      }

      void KeyboardManager::ShowKeyboard()
      {
        assert(window!=0);

        X11::Display *display=dynamic_cast<X11::Display*>(Lum::OS::display);

        if (display==NULL) {
          return;
        }

        ::Window imWindow=GetKeyboardWindow();

        if (imWindow==None) {
          return;
        }

        ::XEvent event;

        memset(&event,0,sizeof(event));
        event.xclient.type=ClientMessage;
        event.xclient.window=imWindow;
        event.xclient.message_type=XInternAtom(display->display,HILDON_IM_ACTIVATE_NAME,False);
        event.xclient.format=HILDON_IM_ACTIVATE_FORMAT;

        HildonIMActivateMessage *msg = reinterpret_cast<HildonIMActivateMessage *>(&event.xclient.data);

        msg->input_window=dynamic_cast<X11::Window*>(window)->GetDrawable();
        msg->app_window=dynamic_cast<X11::Window*>(window)->GetDrawable();
        msg->cmd=HILDON_IM_SETNSHOW;
        //msg->input_mode=HILDON_GTK_INPUT_MODE_FULL;
        msg->trigger=HILDON_IM_TRIGGER_STYLUS;

        if (XSendEvent(display->display,imWindow,False,NoEventMask,&event)==0) {
          std::cerr << "Error while sending OpenKeyboard (_HILDON_IM_ACTIVATE)" << std::endl;
        }

        XSync(display->display,false);
      }

      void KeyboardManager::HideKeyboard()
      {
        X11::Display *display=dynamic_cast<X11::Display*>(Lum::OS::display);

        if (display==NULL) {
          return;
        }

        ::Window imWindow=GetKeyboardWindow();

        if (imWindow==None) {
          return;
        }

        ::XEvent event;

        memset(&event,0,sizeof(event));
        event.xclient.type=ClientMessage;
        event.xclient.window=imWindow;
        event.xclient.message_type=XInternAtom(display->display,HILDON_IM_ACTIVATE_NAME,False);
        event.xclient.format=HILDON_IM_ACTIVATE_FORMAT;

        HildonIMActivateMessage *msg = reinterpret_cast<HildonIMActivateMessage *>(&event.xclient.data);

        msg->input_window=0;
        msg->app_window=0;
        msg->cmd=HILDON_IM_HIDE;
        //msg->input_mode=HILDON_GTK_INPUT_MODE_FULL;
        msg->trigger=HILDON_IM_TRIGGER_STYLUS;

        if (XSendEvent(display->display,imWindow,False,NoEventMask,&event)==0) {
          std::cerr << "Error while sending 'CloseKeyboard' (_HILDON_IM_ACTIVATE)" << std::endl;
        }

        XSync(display->display,false);
      }

      ::Window KeyboardManager::GetKeyboardWindow() const
      {
        X11::Display *display=dynamic_cast<X11::Display*>(Lum::OS::display);

        if (display==NULL) {
          return None;
        }

        Atom hildonWindow=XInternAtom(display->display,"_HILDON_IM_WINDOW",true);

        if (hildonWindow==None) {
          return None;
        }

        Atom           atomType;
        int            atomFormat;
        int            res;
        unsigned long  atomItems;
        unsigned long  atomBytes;
        unsigned char* atomProperties;
        ::Window       window;

        res=XGetWindowProperty(display->display,RootWindow(display->display,display->scrNum),
                               hildonWindow,0,1,false,display->atoms[display->windowAtom],
                               &atomType,&atomFormat,&atomItems,&atomBytes,&atomProperties);
        if (res!=Success ||
            atomType!=XA_WINDOW ||
            atomFormat!=HILDON_IM_WINDOW_ID_FORMAT ||
            atomItems!=1 ||
            atomProperties==NULL) {
          std::cerr << "Cannot get IM window (_HILDON_IM_WINDOW)" << std::endl;
          return None;
        }

        window=reinterpret_cast< ::Window*>(atomProperties)[0];

        XFree(atomProperties);

        return window;
      }

      void KeyboardManager::Cut()
      {
        Lum::Base::DataExchangeObject *object=OS::display->GetSelectionOwner();

        if (object==NULL) {
          return;
        }

        Lum::Base::Data data;

        if (!object->ExecuteAction(Lum::Base::Datatype(Lum::Base::Datatype::text,
                                                       Lum::Base::Datatype::plain),
                                   data,
                                   Lum::Base::DataExchangeObject::actionMove)) {
          return;
        }

        OS::display->SetClipboard(data.GetText());

        object->PostprocessAction(Lum::Base::DataExchangeObject::actionMove);
      }

      void KeyboardManager::Copy()
      {
        Lum::Base::DataExchangeObject *object=OS::display->GetSelectionOwner();

        if (object==NULL) {
          return;
        }

        Lum::Base::Data data;

        if (!object->ExecuteAction(Lum::Base::Datatype(Lum::Base::Datatype::text,
                                                       Lum::Base::Datatype::plain),
                                   data,
                                   Lum::Base::DataExchangeObject::actionCopy)) {
          return;
        }

        OS::display->SetClipboard(data.GetText());
      }

      void KeyboardManager::Paste()
      {
        if (object==NULL) {
          return;
        }

        std::wstring clipboard;

        clipboard=OS::display->GetClipboard();

        if (clipboard.empty()) {
          return;
        }

        Lum::Base::Data data(clipboard);

        object->ExecuteAction(Lum::Base::Datatype(Lum::Base::Datatype::text,
                                                  Lum::Base::Datatype::plain),
                              data,
                              Lum::Base::DataExchangeObject::actionInsert);
      }
    }
  }
}
